#include "ConvexHull.h"
#include <stdlib.h>
#include "MergeSort.h"



ConvexHull::ConvexHull() { };

ConvexHull::~ConvexHull() { };

void ConvexHull::swap(Point2D &a, Point2D &b) {
  Point2D t = a;
  a = b;
  b = t;
};

Polygon* ConvexHull::giftwrapHull(Point2D s[], int n) {
  int a, i;
  for (a = 0, i = 1; i < n; i++) {
    if (s[i] < s[a]) { 
      a = i; 
    };
  };
  s[n] = s[a];
  Polygon *p = new Polygon;
  for (int m = 0; m < n; m++) {
    swap(s[a], s[m]);
    p->insert(s[m]);
    a = m + 1;
    for (int j = m + 2; j <= n; j++) {
      int c = s[j].classify(s[m], s[a]);
      if (c == Point2D::LEFT || c == Point2D::BEYOND) {
        a = j;
      };
    };
    if (a == n) {
      return p;
    };
  };
  return NULL;
};

int leftToRightCmp(Point2D *a, Point2D *b) {
  if (*a < *b) return -1; 
  if (*a > *b) return 1;
  return 0;
};

int rightToLeftCmp(Point2D *a, Point2D *b) {
  return leftToRightCmp(b, a);
};

Vertex* ConvexHull::leastVertex(Polygon &p, int (*cmp)(Point2D*,Point2D*)) {
  Vertex *bestV = p.v();
  p.advance(Vertex::CLOCKWISE);
  for (int i = 1; i < p.size(); p.advance(Vertex::CLOCKWISE), i++)
    if ((*cmp)(p.v(), bestV) < 0) 
      bestV = p.v();
  p.setV(bestV);
  return bestV;
};


Polygon* ConvexHull::mergeHull(Point2D pts[], int n) {
  Point2D **p = new (Point2D*)[n];
  for (int i = 0; i < n; i++)
    p[i] = &pts[i];
  mergeSort(p, n, leftToRightCmp);
  return mHull(p, n);
};

Polygon* ConvexHull::mHull(Point2D *p[], int n) {
  if (n == 1) {
    Polygon *q = new Polygon;
    q->insert(*p[0]);
    return q;
  } else {
    int m = n / 2;
    Polygon *L = mHull(p, m);
    Polygon *R = mHull(p+m, n-m);
    return merge(L, R);
  };
};

Polygon* ConvexHull::merge(Polygon *L, Polygon *R) {
  Vertex *l1, *r1, *l2, *r2;
  Vertex *vl = leastVertex(*L, rightToLeftCmp);
  Vertex *vr = leastVertex(*R, leftToRightCmp);
  bridge(L, R, l1, r1, UPPER);
  L->setV(vl);
  R->setV(vr);
  bridge(L, R, l2, r2, LOWER);
  L->setV(l1);
  L->split(r1);
  R->setV(r2);
  delete R->split(l2);
  return R;
};

void ConvexHull::bridge(Polygon *L, Polygon *R, Vertex* &vl, Vertex* &vr, int type) {
  int sides[2] = { Point2D::LEFT, Point2D::RIGHT };
  int indx = (type == UPPER) ? 0 : 1;
  do {
    vl = L->v();
    vr = R->v();
    supportingLine(L->point(), R, sides[indx]);
    supportingLine(R->point(), L, sides[1-indx]);
  } while ((vl != L->v()) || (vr != R->v()));
};

void ConvexHull::supportingLine(Point2D &s, Polygon *p, int side) {
  int rotation = (side == Point2D::LEFT) ? Vertex::CLOCKWISE : Vertex::COUNTERCLOCKWISE;
  Vertex *a = p->v();
  Vertex *b = p->neighbor(rotation);
  int c = b->classify(s, *a);
  while ((c == side) || (c == Point2D::BEYOND) || (c == Point2D::BETWEEN)) {
    p->advance(rotation);
    a = p->v();
    b = p->neighbor(rotation);
    c = b->classify(s, *a);
  };
};


Point2D somePoint;   // globally used for Cmp-Routine

int closestToPolygonCmp(Point2D *a, Point2D *b) {
  double distA = (somePoint - *a).length();
  double distB = (somePoint - *b).length();
  if (distA < distB) return -1;
  else if (distA > distB) return 1;
  return 0;
};

enum { INSIDE, OUTSIDE, BOUNDARY };
enum { TOUCHING, CROSSING, INESSENTIAL };

int edgeType(Point2D &a, Edge &e) {
  Point2D v = e.org;
  Point2D w = e.dest;
  switch(a.classify(e)) {
  case Point2D::LEFT:
    return ((v.y<a.y) && (a.y<=w.x)) ? CROSSING : INESSENTIAL;
  case Point2D::RIGHT:
    return ((w.y<a.y) && (a.y<=v.y)) ? CROSSING : INESSENTIAL;
  case Point2D::BETWEEN:
  case Point2D::ORIGIN:
  case Point2D::DESTINATION:
    return TOUCHING;
  default:
    return INESSENTIAL;
  };
};
 
int pointInPolygon(Point2D &a, Polygon &p) {
  int parity = 0;
  for(int i = 0; i < p.size(); i++, p.advance(Vertex::CLOCKWISE)) {
    Edge e = p.edge();
    switch (edgeType(a, e)) {
    case TOUCHING:
      return BOUNDARY;
    case CROSSING:
      parity = 1 - parity;
    };
  };
  return (parity ? INSIDE : OUTSIDE);
};

int pointInConvexPolygon(Point2D &s, Polygon &p) {
  if (p.size() == 1) 
    return (s == p.point());
  if (p.size() == 2) {
    int c = s.classify(p.edge());
    return ((c==Point2D::BETWEEN) || (c==Point2D::ORIGIN) || (c==Point2D::DESTINATION));
  };
  Vertex *org = p.v();
  for (int i =0; i < p.size(); i++, p.advance(Vertex::CLOCKWISE))
    if (s.classify(p.edge()) == Point2D::LEFT) {
      p.setV(org);
      return false;
    };
  return true;
};

Polygon* ConvexHull::insertionHull(Point2D s[], int n) {
  Polygon *p = new Polygon;
  p->insert(s[0]);
  for (int i = 1; i < n; i++) {
    if (pointInConvexPolygon(s[i], *p))
      continue;
    somePoint = s[i];
    leastVertex(*p, closestToPolygonCmp);
    supportingLine(s[i], p, Point2D::LEFT);
    Vertex *l = p->v();
    supportingLine(s[i], p, Point2D::RIGHT);
    delete p->split(l);
    p->insert(s[i]);
  };
  return p;
};


int ConvexHull::polarCmp(Point2D *p, Point2D *q, Point2D *o) {
  Point2D vp = *p - *o;
  Point2D vq = *q - *o;
  double pPolar = vp.polarAngle();
  double qPolar = vq.polarAngle();
  if (pPolar < qPolar) return -1;
  if (pPolar > qPolar) return 1;
  if (vp.length() < vq.length()) return -1;
  if (vp.length() > vq.length()) return 1;
  return 0;
};

