/*
 * Geometry2D.C 
 *
 * Copyright (C) 1994-96, Stephan Amann <amann@iam.unibe.ch>
 *                        Christoph Streit <streit@iam.unibe.ch>
 *                        University of Berne, Switzerland
 *
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 * -----------------------------------------------------------------------------
 *  $Id: Geometry2D.C,v 1.8 1996/05/29 13:32:44 streit Exp $
 * -----------------------------------------------------------------------------
 */
 
#include <strstream.h>
#include "tess-interface.h"
#include "Geometry2D.h"

//___________________________________________________________________ Geometry2D

Geometry2D::OrientType Geometry2D::orient(Real px1, Real py1,
                                          Real px2, Real py2,
                                          Real px3, Real py3)
{
  Real dx1 = px2 - px1;
  Real dy1 = py2 - py1;
  Real dx2 = px3 - px1; 
  Real dy2 = py3 - py1;
  Real dx1dy2 = dx1*dy2;
  Real dy1dx2 = dy1*dx2;
 
  if (dx1dy2 > dy1dx2) return CCW; // positive sign
  if (dx1dy2 < dy1dx2) return CW;  // negative sign
  if (dx1*dx2 < 0 || dy1*dy2 < 0) return CW; 
  if ((dx1*dx1 + dy1*dy1) >= (dx2*dx2 + dy2*dy2)) return ZERO; // area is 0
  
  return CCW;
}

bool Geometry2D::intersect(Real px1, Real py1,
                           Real px2, Real py2,
                           Real px3, Real py3,
                           Real px4, Real py4)
{
  if (orient(px1,py1, px2,py2, px3,py3)*orient(px1,py1, px2,py2, px4,py4) <= 0 &&
      orient(px3,py3, px4,py4, px1,py1)*orient(px3,py3, px4,py4, px2,py2) <= 0) {
    Real denom = (px3-px4)*py1 + (px4-px3)*py2 +
                 (px2-px1)*py3 + (px1-px2)*py4;
    
    // Better test overlap.
    if (fabs(denom) < EPSILON)
      return true;             
      
    // No intersection if the segments only touch.
    if ((equal(px1, px3) && equal(py1, py3)) || 
        (equal(px1, px4) && equal(py1, py4)) ||
        (equal(px2, px3) && equal(py2, py3)) ||
        (equal(px2, px4) && equal(py2, py4)))
      return false;             

    return true;
  }
  
  return false;
}
 
bool Geometry2D::pointInTriangle(Real px,  Real py,
                                 Real px1, Real py1,
                                 Real px2, Real py2,
                                 Real px3, Real py3)
{
  if (orient(px1,py1, px2,py2, px, py) == CCW &&
      orient(px2,py2, px3,py3, px, py) == CCW &&
      orient(px3,py3, px1,py1, px, py) == CCW)
    return true;
  return false;
}

List<Geometry2D::Triangle>* 
Geometry2D::createPolygonTriangulation(List<Vector2D>& vertices)
{
  long numVertices = vertices.count();
  register long i;

  if (numVertices < 3) {
    ostrstream os;
    os << "[Geometry2D::createPolygonTriangulation] only " << numVertices << " provided"
       << " , >= 3 required";
    Report::recoverable(os); 
    return NULL;
  }

  //
  // Check for degenerated polygon (straight line...)
  //
  Real area = 0;
  Vector2D v1, v2;

  for (i=1; i<numVertices-1; i++) {
    v1 = vertices.item(i-1) - vertices.item(i);
    v2 = vertices.item(i+1) - vertices.item(i);
    area += fabs(v1.x()*v2.y()-v1.y()*v2.x());
  }

  if (equal(area, 0))
    return NULL;

  //
  // Start tesselation.
  //
  List<Geometry2D::Triangle>* result = new List<Geometry2D::Triangle>(vertices.count()-2);

  int cntr[] = { (int)numVertices };

  double (*verticesArray)[2] = (double(*)[2]) new double[2*(numVertices+1)];

  for (i=0; i<numVertices; i++) {
    verticesArray[i+1][0] = vertices.item(i).x();
    verticesArray[i+1][1] = vertices.item(i).y();
  }

  int (*triangles)[3] = (int(*)[3])new int[3*numVertices];

  int numTriangles = triangulate_polygon(1, cntr, verticesArray, triangles);
    
  for (i = 0; i < numTriangles; i++)
    result->append(Triangle(triangles[i][0]-1, triangles[i][1]-1, triangles[i][2]-1));
  
  delete [] verticesArray;
  delete [] triangles;

  return result;
}
