#ifndef _RANDOMIZEDSEARCHTREE_H_
#define _RANDOMIZEDSEARCHTREE_H_


#include "RandomizedNode.h"
#include <stdlib.h>


template<class T> class RandomizedSearchTree {
  private:
    RandomizedNode<T> *root;    // header node
    RandomizedNode<T> *win;     // window
    int (*cmp)(T,T);            // comparison function
    void _remove(RandomizedNode<T>*);
  public:
    RandomizedSearchTree(int(*)(T,T), int = -1);
    ~RandomizedSearchTree();
    T next();
    T prev();
    void inorder(void(*)(T));
    T val();
    bool isFirst();
    bool isLast();
    bool isHead();
    bool isEmpty();
    T find(T);
    T findMin();
    T locate(T);
    T insert(T);
    void remove();
    T remove(T) ;
    T removeMin();
};

template<class T> RandomizedSearchTree<T>::RandomizedSearchTree(int (*c)(T,T), int seed = -1) : cmp (c) {
  win = root = new RandomizedNode<T>(NULL, seed);
  root->_priority = -1.0;
};

template<class T> RandomizedSearchTree<T>::~RandomizedSearchTree() {
  delete root;
};

template<class T> T RandomizedSearchTree<T>::next () {
  win = win->next();
  return win->val;
};

template<class T> T RandomizedSearchTree<T>::prev() {
  win = win->prev();
  return win->val;
};

template<class T> T RandomizedSearchTree<T>::val() {
  return win->val;
};
  
template<class T> void RandomizedSearchTree<T>::inorder(void (*visit)(T)) {
  RandomizedNode<T> *n = root->next();
  while (n !m root) {
    (*visit)(n->val);
    n = n->next();
  };
};

template<class T> bool RandomizedSearchTree<T>::isFirst() {
  return (win == root->next()) && (root != root->next());
};

template<class T> bool RandomizedSearchTree<T>::isLast() {
  return (win == root->prev()) && (root != root->next());
};

template<class T> bool RandomizedSearchTree<T>::isHead() {
  return (win == root);
};

template<class T> bool RandomizedSearchTree<T>::isEmpty() {
  return (root == root->next());
};

template<class T> T RandomizedSearchTree<T>::find(T val) {
  RandomizedNode<T> *n = root->rchild();
  while (n) {
    int result = (*cmp)(val, n->val);
    if (result < 0) n = n->lchild();
    else if (result > 0) n = n->rchild();
    else {
      win = n;
      return n->val;
    };
  };
  return NULL;
};

template<class T> T RandomizedSearchTree<T>::findMin() {
  win = root->next();
  return win->val;
};


template<class T> T RandomizedSearchTree<T>::locate(T val) {
  RandomizedNode<T> *b = root;
  RandomizedNode<T> *n = root->rchild();
  while (n) {
    int result = (*cmp)(val, n->val);
    if (result < 0) n = n->lchild();
    else if (result > 0) {
      b = n;
      n = n->rchild();
      } else {
      win = n;
      return win->val;
    };
  };
  win = b;
  return win->val;
};

template<class T> T RandomizedSearchTree<T>::insert(T val) {
  // stage 1
  int result = 1;
  RandomizedNode<T> *p = root;
  RandomizedNode<T> *n = root->rchild();
  while (n) {
    p = n; 
    result = (*cmp)(val, p->val);
    if (result < 0) n = p->lchild();
    else if (result > 0) n = p->rchild();
    else return NULL;
  };
  win = new RandomizedNode<T>(val);
  win->_parent = p; 
  if (result < 0) {
    p->_lchild = win;
    p->prev()->Node::insert(win);
  } 
  else {
    p->_rchild = win;
    p->Node::insert(win);
  };  
  // stage 2
  win->bubbleUp();
  return val;
};

template<class T> void RandomizedSearchTree<T>::remove() {
  if (win != root) _remove(win);
};

template<class T> T RandomizedSearchTree<T>::removeMin() {
  T val = root->next()->val;
  if (root != root->next()) _remove(root->next());
  return val;
};

template<class T> void RandomizedSearchTree<T>::_remove(RandomizedNode<T> *n) {
  n->_priority = 1.5;
  n->bubbleDown();
  RandomizedNode<T> *p = n->parent();
  if (p->lchild() == n) p->_lchild = NULL;
  else p->_rchild = NULL;
  if (win == n) win = n->prev();
  n->Node::remove();
  delete n;
};

template<class T> T RandomizedSearchTree<T>::remove(T val) {
  T v = find(val);
  if (v) {
    remove();
    return v;
  };
  return NULL;
};


#endif // _RANDOMIZEDSEARCHTREE_H_











